/******************************************************************************
 * $Id: mappdf.c 7585 2008-05-13 05:19:24Z tbonfort $
 *
 * Project:  MapServer
 * Purpose:  An api for PDF format output using the pdflib or libharu library
 *           http://www.pdflib.com
 *           http://libharu.org
 * Author:   <jwall@webpeak.com> , <jspielberg@webpeak.com>
 * libHaru support: <kyngchaos@kyngchaos.com>
 *
 ******************************************************************************
 * Copyright (c) 1996-2005 Regents of the University of Minnesota.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies of this Software or works derived from this Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ****************************************************************************/

#ifdef USE_PDF

#include <assert.h>
#if !defined(_WIN32)
#include <zlib.h>
#endif
#include "mapserver.h"
#if defined USE_HARUPDF
#include <math.h>
#endif

MS_CVSID("$Id: mappdf.c 7585 2008-05-13 05:19:24Z tbonfort $")

/* -------------------------------------------------------------------- */
/*      prototypes.                                                     */
/* -------------------------------------------------------------------- */
/*see mapserver.h*/

#include "mappdf.h"

/* -------------------------------------------------------------------- */
/*      PDF API functions and aliases                                   */
/* -------------------------------------------------------------------- */

/* NOTE: PDFlib PDF object is a pointer, thus needs a * def.  libHaru   */
/*       PDF, page and other objects are structs, and don't need * def. */

/* error handler for libharu */
#if defined USE_HARUPDF
static void error_handler (HPDF_STATUS error_no, HPDF_STATUS detail_no, void *user_data)
{
    /* from haru_font_demo.c:
    printf ("ERROR: error_no=%04X, detail_no=%u\n", (HPDF_UINT)error_no,
                (HPDF_UINT)detail_no);
    longjmp(env, 1);
    */
    msSetError(MS_GDERR, "HPDF error=%04X , detail=%u.", "HPDF_Doc",
                   (HPDF_UINT)error_no, (HPDF_UINT)detail_no);
}
#endif

/* translate page matrix */
static void MAPPDF_Translate(MAPPDF_DRAWOBJ_TYPE pdfpage, float x, float y)
{
#if defined USE_HARUPDF
    HPDF_Page_Concat(pdfpage, 1, 0, 0, 1, x, y);
#elif defined USE_PDFLIB
    PDF_translate(pdfpage, x, y);
#endif
}

/* scale page matrix */
static void MAPPDF_Scale(MAPPDF_DRAWOBJ_TYPE pdfpage, float sx, float sy)
{
#if defined USE_HARUPDF
    HPDF_Page_Concat(pdfpage, sx, 0, 0, sy, 0, 0);
#elif defined USE_PDFLIB
    PDF_scale(pdfpage, sx, sy);
#endif
}

/* rotate page matrix */
static void MAPPDF_Rotate(MAPPDF_DRAWOBJ_TYPE pdfpage, float a)
{
#if defined USE_HARUPDF
    float ca = cosf(MS_DEG_TO_RAD * a);
    float sa = sinf(MS_DEG_TO_RAD * a);
    HPDF_Page_Concat(pdfpage, ca, sa, -sa, ca, 0, 0);
#elif defined USE_PDFLIB
    PDF_rotate(pdfpage, a);
#endif
}

/************************************************************************/
/*                        PDF *initializeDocument                       */
/*                                                                      */
/*              Internal function to create a PDF type                  */
/************************************************************************/
MAPPDF_DOC_TYPE initializeDocument()
{
    /* First we need to make the PDF */
#if defined USE_HARUPDF
    /* Haru always starts in memory */
    HPDF_Doc pdf = HPDF_New(error_handler, NULL);
#elif defined USE_PDFLIB
    PDF *pdf = PDF_new();
    PDF_open_file (pdf,""); /*This creates the file in-memory, not on disk */
#endif
    MAPPDF_FN_SETINFO (pdf, MAPPDF_INFO_CREATOR, "UMN MapServer");
    MAPPDF_FN_SETINFO (pdf, MAPPDF_INFO_AUTHOR, "Implementation for MapServer 3.7 - DM Solutions Group, based on original work by Market Insite Group");
    MAPPDF_FN_SETINFO (pdf, MAPPDF_INFO_TITLE, "DM Solutions Group and Market Insite PDF Map");

    return pdf;
}

/************************************************************************/
/*int msLoadFontSetPDF                                                  */
/*                                                                      */
/* load fonts from file.  for libHaru, we need to setup extra hash      */
/* table to track font name aliases.                                    */
/************************************************************************/
/**** FIXME  loading ALL fontset fonts into the PDF is silly, they may not
             all be used in the map file.  Improved font loading: load
             as needed in the text drawing functions (orig PDFlib
             sortof does this already.  libHaru still needs to track
             alias names.
             Or, maybe PDFlib and Haru clean out unused fonts on save? ****/

#if defined USE_HARUPDF
int msLoadFontSetPDF(fontSetObj *fontset, HPDF_Doc pdf, hashTableObj *hfonts)
#elif defined USE_PDFLIB
int msLoadFontSetPDF(fontSetObj *fontset, PDF *pdf)
#endif
{

#if defined USE_HARUPDF
    HPDF_Page   page;
#endif
    FILE        *stream;
    char        buffer[MS_BUFFER_LENGTH];
    char        alias[64], file1[MS_PATH_LENGTH], file2[MS_PATH_LENGTH];
    char        *path=NULL, *fullPath, szPath[MS_MAXPATHLEN];
    int         i;

    if (fontset == NULL) return(0);
    if (fontset->filename == NULL) return(0);
    
#if defined USE_HARUPDF
    page = HPDF_GetCurrentPage(pdf);
    if (!page)
    {
        msSetError(MS_GDERR, "Unable to get current pdf page.", "msLoadFontSetPDF()");
        return(-1);
    }
    /*initHashTable(hfonts);*/
    hfonts = msCreateHashTable();
#endif

    /**** FIXME  why not use the path paths from the fontset object? - 
                 hasn't the fontset data already been loaded? ****/
    stream =
      fopen(msBuildPath(szPath, fontset->map->mappath, fontset->filename), "r");

    if (!stream)
    {
        msSetError(MS_IOERR, "Error opening fontset %s.", "msLoadFontSetPDF()",
                   fontset->filename);
        return(-1);
    }

    i = 0;
    path = msGetPath(fontset->filename);

    while (fgets(buffer, MS_BUFFER_LENGTH, stream))
    { /* while there's something to load */
#if defined USE_HARUPDF
        /* get discard qualifiers warning otherwise for haru */
        const
#endif
              char *fontParam;

        if(buffer[0] == '#' ||
           buffer[0] == '\n' ||
           buffer[0] == '\r' ||
           buffer[0] == ' ')
            continue; /* skip comments and blank lines */

        fullPath = NULL;
        sscanf(buffer, "%s %s", alias,  file1);

        fullPath = file1;

#if defined(_WIN32) && !defined(__CYGWIN__)
        if(strlen(file1) <= 1 || (file1[0] != '\\' && file1[1] != ':'))
        { /* not a full path */
            sprintf(file2, "%s%s", path, file1);
            fullPath = msBuildPath(szPath, fontset->map->mappath, file2);
        }
#else
       if(file1[0] != '/')
        { /* not full path */
            sprintf(file2, "%s%s", path, file1);
            fullPath = msBuildPath(szPath, fontset->map->mappath, file2);
        }
#endif

        /*ok we have the alias and the full name*/
#if defined USE_HARUPDF
        /* for haru we need to load font AND store returned name in a hashtable */
        /**** FIXME  assume TT only for now, need to check file extension
                     and use appropriate load function ****/
        fontParam = HPDF_LoadTTFontFromFile(pdf, fullPath, 1);
        msInsertHashTable(hfonts, alias, fontParam);
#elif defined USE_PDFLIB
        /* PDFlib allows alias naming of fonts, so it keeps track for us */
        fontParam = (char *)malloc(sizeof(char)*(strlen(fullPath)+strlen(alias)+3));
        sprintf(fontParam, "%s==%s", alias, fullPath);
        PDF_set_parameter(pdf, "FontOutline", fontParam);
        free(fontParam);
#endif
        i++;
    }

    fclose(stream); /* close the file */

    if( path )
        free(path);

    return(0);
}

/************************************************************************/
/*                        imageObj *msImageCreatePDF                    */
/*                                                                      */
/*      Utility function to create an image object of PDF type          */
/************************************************************************/
imageObj *msImageCreatePDF(int width, int height, outputFormatObj *format,
                           char *imagepath, char *imageurl, mapObj *map)
{
    imageObj         *oImage = NULL;
    MAPPDF_DOC_TYPE  pdf = NULL;
#if defined USE_HARUPDF
    HPDF_Page        page = NULL;
#endif
    char             *driver = strdup("GD/GIF");

    assert(strcasecmp(format->driver,"PDF") == 0);

    oImage = (imageObj *)calloc(1,sizeof(imageObj));

    oImage->format = format;
    format->refcount++;

    oImage->width = width;
    oImage->height = height;
    oImage->imagepath = NULL;
    oImage->imageurl = NULL;

    if (imagepath)
    {
        oImage->imagepath = strdup(imagepath);
    }
    if (imageurl)
    {
        oImage->imageurl = strdup(imageurl);
    }

    oImage->img.pdf = (PDFObj *)malloc(sizeof(PDFObj));
    oImage->img.pdf->pdf = NULL;
    oImage->img.pdf->pdf = initializeDocument();
    oImage->img.pdf->map = map;
    if(!oImage->img.pdf->pdf)
    {
        msSetError(MS_GDERR, "Unable to initialize pdfPage.", "msDrawMap()");
        return(NULL);
    }

    pdf = oImage->img.pdf->pdf;
    /*  load fonts and set the pdf ready to be drawn into*/
#if defined USE_HARUPDF
    /* oImage->img.pdf->hfonts = NULL; */
    msLoadFontSetPDF(&(map->fontset), pdf, &(oImage->img.pdf->hfonts));
    page = HPDF_AddPage(pdf);
    if (!page)
    {
        msSetError(MS_GDERR, "Unable to add new pdf page.", "msDrawMap()");
        return(NULL);
    }
    HPDF_Page_SetWidth(page, (float)width);
    HPDF_Page_SetHeight(page, (float)height);
#elif defined USE_PDFLIB
    msLoadFontSetPDF((&(map->fontset)), pdf);
    PDF_begin_page(pdf, (float)width, (float)height);
#endif

    /* pdf has it's origin in the bottom left so we need to flip the coordinate */
    /* system to fit with the rest of mapserver. */
    MAPPDF_Translate(MAPPDF_DRAWOBJ, 0, (float)map->height);
    MAPPDF_Scale(MAPPDF_DRAWOBJ, 1, -1);

    /*render text with both a fill and a stroke*/
#if defined USE_HARUPDF
    HPDF_Page_SetTextRenderingMode(page, HPDF_FILL_THEN_STROKE);
#elif defined USE_PDFLIB
    PDF_set_value(pdf, "textrendering", 2);
#endif

/* -------------------------------------------------------------------- */
/*      if the output raster, we create a GD image that                 */
/*      will be used to contain the rendering of all the layers.        */
/* -------------------------------------------------------------------- */
    if (strcasecmp(msGetOutputFormatOption(oImage->format,"OUTPUT_TYPE",""), 
                   "RASTER") != 0)
    {
        oImage->img.pdf->imagetmp = NULL;
    }
    else
    {
#ifdef USE_GD_GIF
        driver = strdup("GD/GIF");
#else

#ifdef USE_GD_PNG
        driver = strdup("GD/PNG");
#else

#ifdef USE_GD_JPEG
        driver = strdup("GD/JPEG");
#else

#ifdef USE_GD_WBMP
        driver = strdup("GD/WBMP");
#endif

#endif
#endif
#endif

        oImage->img.pdf->imagetmp = (imageObj *) 
          msImageCreateGD(map->width, map->height,
                          msCreateDefaultOutputFormat(map, driver),
                          map->web.imagepath, map->web.imageurl);
    }
    return oImage;
}

/************************************************************************/
/* void drawPolylinePDF                                                 */
/*                                                                      */
/* internal function to do line drawing in the pdf                      */
/************************************************************************/
void drawPolylinePDF(MAPPDF_DOC_TYPE pdf, shapeObj *p, colorObj  *c, double width)
{
    int        i, j;
#if defined USE_HARUPDF
    HPDF_Page  page = NULL;

    page = HPDF_GetCurrentPage(pdf);
    if (!page)
    {
        msSetError(MS_GDERR, "Unable to get current pdf page.", "msDrawMap()");
        return;
    }
#endif

    if (width)
    {
        MAPPDF_FN_SETLINEJOIN(MAPPDF_DRAWOBJ, MAPPDF_LINEJOIN_ROUND);
        MAPPDF_FN_SETLINEWIDTH(MAPPDF_DRAWOBJ, (float)width);
    }

    if (c)
    {
        /* old version set both fill & stroke color, but we only need stroke */
        MAPPDF_FN_SETRGBSTROKECOLOR(MAPPDF_DRAWOBJ, (float)c->red/255,
                                                    (float)c->green/255,
                                                    (float)c->blue/255);
    }

    for (i = 0; i < p->numlines; i++)
    {
        if (p->line[i].numpoints)
        {
            MAPPDF_FN_MOVETO(MAPPDF_DRAWOBJ, (float)p->line[i].point[0].x,
                                             (float)p->line[i].point[0].y);
        }
        for(j=1; j<p->line[i].numpoints; j++)
        {
            MAPPDF_FN_LINETO(MAPPDF_DRAWOBJ, (float)p->line[i].point[j].x,
                                             (float)p->line[i].point[j].y);
        }
    }
    MAPPDF_FN_STROKE(MAPPDF_DRAWOBJ);
/*  MAPPDF_FN_SETLINEJOIN(MAPPDF_DRAWOBJ, MAPPDF_LINEJOIN_MITER); */
    MAPPDF_FN_SETLINEWIDTH(MAPPDF_DRAWOBJ, 1);

}

/**** FIXME  why pdflib >= 6?  I see dashed lines in v5.  bugs? ****/
#if defined USE_HARUPDF || (defined USE_PDFLIB && PDFLIB_MAJORVERSION >= 6)
 /************************************************************************/
 /* void drawDashedPolylinePDF                                           */
 /*                                                                      */
 /* internal function to do line drawing in the pdf, using dashed        */
 /* patterns.  Note that haru currently has a max pattern length of 8    */
 /* while MS has a max of 10.                                            */
 /************************************************************************/
 void drawDashedPolylinePDF(MAPPDF_DOC_TYPE pdf, shapeObj *p, symbolObj  *s,
                            colorObj *c, double width)
 {
    int          i, j;
#if defined USE_HARUPDF
    HPDF_Page    page = NULL;
    HPDF_UINT16  optlist[8];
#elif defined USE_PDFLIB
    int          optlistlength = 14;
    char         *optlist = NULL;
#endif

#if defined USE_HARUPDF
    page = HPDF_GetCurrentPage(pdf);
    if (!page)
    {
        msSetError(MS_GDERR, "Unable to get current pdf page.", "msDrawMap()");
        return;
    }

    if (s->patternlength > 8)
    {
        msSetError(MS_SYMERR, "Max PDF pattern length is 8.", "drawDashedPolylinePDF()");
        return;
    }
#endif
    
    if (s->patternlength > 0)
    {
        for (i = 0; i < s->patternlength; i++)
        {
            if (s->pattern[i] > 0)
#if defined USE_HARUPDF
                optlist[i] = (HPDF_UINT16)s->pattern[i];
#elif defined USE_PDFLIB
                optlistlength += (int)(log10(s->pattern[i])) + 2;
#endif
            else
            {
                msSetError(MS_SYMERR, "Symbol patterns must be positive.", "drawDashedPolylinePDF()");
                return;
            }
        }
#if defined USE_HARUPDF
        HPDF_Page_SetDash(page, optlist, s->patternlength, 0);
#elif defined USE_PDFLIB
        optlist = (char*)malloc(optlistlength*sizeof(char));
        sprintf(optlist, "dasharray={");
        for (i = 0; i < s->patternlength; i++)
            sprintf(optlist, "%s %d", optlist, s->pattern[i]);
        sprintf(optlist, "%s }", optlist);
        PDF_setdashpattern(pdf, optlist);
#endif
    }

    if (width)
    {
        MAPPDF_FN_SETLINEJOIN(MAPPDF_DRAWOBJ, MAPPDF_LINEJOIN_ROUND);
        MAPPDF_FN_SETLINEWIDTH(MAPPDF_DRAWOBJ, (float)width);
    }

    if (c)
    {
        MAPPDF_FN_SETRGBSTROKECOLOR(MAPPDF_DRAWOBJ, (float)c->red/255,
                                                    (float)c->green/255,
                                                    (float)c->blue/255);
    }

    for (i = 0; i < p->numlines; i++)
    {
        if (p->line[i].numpoints)
        {
            MAPPDF_FN_MOVETO(MAPPDF_DRAWOBJ, (float)p->line[i].point[0].x,
                                             (float)p->line[i].point[0].y);
        }
        for(j = 1; j<p->line[i].numpoints; j++)
        {
            MAPPDF_FN_LINETO(MAPPDF_DRAWOBJ, (float)p->line[i].point[j].x,
                                             (float)p->line[i].point[j].y);
        }
    }
     
    MAPPDF_FN_STROKE(MAPPDF_DRAWOBJ);
#if defined USE_HARUPDF
    HPDF_Page_SetDash(page, NULL, 0, 0);
#elif defined USE_PDFLIB
    PDF_setdashpattern(pdf, "");
    free(optlist); 
#endif
    MAPPDF_FN_SETLINEWIDTH(MAPPDF_DRAWOBJ, 1);
}
#endif

/************************************************************************/
/*  void msDrawLineSymbolPDF                                            */
/*                                                                      */
/*  Draw a line symbol of the specified size and color                  */
/************************************************************************/
void msDrawLineSymbolPDF(symbolSetObj *symbolset, imageObj *image, shapeObj *p,
                         styleObj *style, double scalefactor)
{
    int              size;
    MAPPDF_DOC_TYPE  pdf;

    if (style->size == -1)
    {
        size = msSymbolGetDefaultSize(symbolset->symbol[style->symbol]);
        size = MS_NINT(size*scalefactor);
    }
    else
        size = MS_NINT(style->size*scalefactor);
    pdf = image->img.pdf->pdf;

    if (p->numlines <= 0)
        return;

  /* no such symbol, 0 is OK */
    if (style->symbol > symbolset->numsymbols || style->symbol < 0)
        return;

    if (!(MS_VALID_COLOR(style->color))) /* invalid color */
        return;

    if (size < 1) /* size too small */
        return;

#if defined USE_HARUPDF || (defined USE_PDFLIB && PDFLIB_MAJORVERSION >= 6)
    if (symbolset && symbolset->symbol[style->symbol]->patternlength)
       drawDashedPolylinePDF(pdf, p, symbolset->symbol[style->symbol], 
                             &(style->color), size);
    else
#endif
       drawPolylinePDF(pdf, p, &(style->color),  size);

    return;
}

/************************************************************************/
/*  void msFilledPolygonPDF                                             */
/*                                                                      */
/*  creates a filled polygon in the pdf from a shape                    */
/*  Note: not part of external api, despite name                        */
/************************************************************************/
void msFilledPolygonPDF(MAPPDF_DOC_TYPE pdf, shapeObj *p, colorObj  *c)
{
    int        i, j;
#if defined USE_HARUPDF
    HPDF_Page  page = NULL;

    page = HPDF_GetCurrentPage(pdf);
    if (!page)
    {
        msSetError(MS_GDERR, "Unable to get current pdf page.", "msDrawMap()");
        return;
    }
#endif

    if (c)
    {
        MAPPDF_FN_SETRGBSTROKECOLOR(MAPPDF_DRAWOBJ, (float)c->red/255,
                                                    (float)c->green/255,
                                                    (float)c->blue/255);
        MAPPDF_FN_SETRGBFILLCOLOR(MAPPDF_DRAWOBJ, (float)c->red/255,
                                                  (float)c->green/255,
                                                  (float)c->blue/255);
    }
    
    for (i = 0; i < p->numlines; i++)
    {
        if (p->line[i].numpoints)
        {
            MAPPDF_FN_MOVETO(MAPPDF_DRAWOBJ, (float)p->line[i].point[0].x,
                                             (float)p->line[i].point[0].y);
        }
        for(j=1; j<p->line[i].numpoints; j++){
            MAPPDF_FN_LINETO(MAPPDF_DRAWOBJ, (float)p->line[i].point[j].x,
                                             (float)p->line[i].point[j].y);
        }
    }
    MAPPDF_FN_CLOSEFILLSTROKE(MAPPDF_DRAWOBJ);

}

/************************************************************************/
/*  void billboardPDF                                                   */
/*                                                                      */
/*  creates a filled polygon in the pdf as a background                 */
/************************************************************************/
void billboardPDF(MAPPDF_DOC_TYPE pdf, shapeObj *shape, labelObj *label)
{
    int       i;
    shapeObj  temp;

    msInitShape(&temp);
    msAddLine(&temp, &shape->line[0]);

    if (label->backgroundshadowcolor.pen >= 0)
    {
        for (i=0; i<temp.line[0].numpoints; i++)
        {
            temp.line[0].point[i].x += label->backgroundshadowsizex;
            temp.line[0].point[i].y += label->backgroundshadowsizey;
        }
        msFilledPolygonPDF(pdf, &temp, &label->backgroundshadowcolor);
        for (i=0; i<temp.line[0].numpoints; i++)
        {
            temp.line[0].point[i].x -= label->backgroundshadowsizex;
            temp.line[0].point[i].y -= label->backgroundshadowsizey;
        }
    }

    msFilledPolygonPDF(pdf, &temp, &label->backgroundcolor);

    msFreeShape(&temp);
}

/************************************************************************/
/*  void msDrawLabelPDF                                                 */
/*                                                                      */
/*  draws a single label at the specified position                      */
/************************************************************************/
int msDrawLabelPDF(imageObj *image, pointObj labelPnt, char *string,
                   labelObj *label, fontSetObj *fontset, double scalefactor)
{
    if (!string)
        return(0); /* not an error, just don't want to do anything */

    if (strlen(string) == 0)
        return(0); /* not an error, just don't want to do anything */

    if (label->position != MS_XY)
    {
        pointObj p;
        rectObj r;

        if (msGetLabelSize(image,string, label, &r, fontset, scalefactor, MS_FALSE) == -1)
            return(-1);
        p = get_metrics(&labelPnt, label->position, r,
                                   label->offsetx, label->offsety,
                                   label->angle, 0, NULL);
        msDrawTextPDF(image, p, string,
                 label, fontset, scalefactor);
    }
    else
    {
        labelPnt.x += label->offsetx;
        labelPnt.y += label->offsety;
        msDrawTextPDF(image,labelPnt, string,
                 label, fontset, scalefactor);
    }

    return(0);

}

/************************************************************************/
/*  int msDrawLabelCachePDF                                             */
/*                                                                      */
/*  draws a set of labels                                               */
/************************************************************************/
int msDrawLabelCachePDF(imageObj *image, mapObj *map)
{
    pointObj             p;
    int                  i, j, l, priority;
    rectObj              r;
    labelCacheMemberObj  *cachePtr=NULL;
    layerObj             *layerPtr=NULL;
    labelObj             *labelPtr=NULL;

    int      draw_marker;
    int      marker_width, marker_height;
    int      marker_offset_x, marker_offset_y;
    rectObj  marker_rect;
    
    imageObj *imagetmp = NULL;

/* -------------------------------------------------------------------- */
/*      if not PDF, return.                                             */
/* -------------------------------------------------------------------- */
    if (image == NULL || map == NULL || !MS_DRIVER_PDF(image->format))
        return -1;
/* -------------------------------------------------------------------- */
/*      if the output raster, we crate a GD image that                  */
/*      will be used to conating the rendering of all the layers.       */
/* -------------------------------------------------------------------- */
    if (strcasecmp(msGetOutputFormatOption(image->format,"OUTPUT_TYPE",""), 
                   "RASTER") == 0)
    {
        int orig_renderer;
        imagetmp = (imageObj *)image->img.pdf->imagetmp;
        msImageInitGD( imagetmp, &map->imagecolor);
        orig_renderer=image->format->renderer;
        image->format->renderer = MS_RENDER_WITH_GD;
        msDrawLabelCache(imagetmp, map);
        image->format->renderer = orig_renderer;
        return 0;
    }

    for (priority=MS_MAX_LABEL_PRIORITY-1; priority>=0; priority--)
    {
        labelCacheSlotObj *cacheslot;
        cacheslot = &(map->labelcache.slots[priority]);

        for (l=cacheslot->numlabels-1; l>=0; l--)
        {
            /* point to right spot in cache */
            cachePtr = &(cacheslot->labels[l]);

            /* set a couple of other pointers, avoids nasty references */
            layerPtr = (GET_LAYER(map, cachePtr->layerindex));
            /* classPtr = &(cachePtr->class); */
            labelPtr = &(cachePtr->label);

            if (!cachePtr->text)
                continue; /* not an error, just don't want to do anything */

            if (strlen(cachePtr->text) == 0)
                continue; /* not an error, just don't want to do anything */

            if (msGetLabelSize(image,cachePtr->text, labelPtr, &r, &(map->fontset), layerPtr->scalefactor, MS_TRUE) == -1)
                return(-1);

            if (labelPtr->autominfeaturesize && ((r.maxx-r.minx) > cachePtr->featuresize))
                continue; /* label too large relative to the feature */

            draw_marker = marker_offset_x = marker_offset_y = 0; /* assume no marker */
            if (layerPtr->type == MS_LAYER_ANNOTATION || layerPtr->type == MS_LAYER_POINT)
            { /* there *is* a marker */

	            /* TO DO: at the moment only checks the bottom style,
	               perhaps should check all of them */
                if (cachePtr->numstyles==0)
                    marker_width = marker_height = 1;
                else if (msGetMarkerSize(&map->symbolset, &(cachePtr->styles[0]), &marker_width, &marker_height, layerPtr->scalefactor) != MS_SUCCESS)
	                return(-1);

                marker_offset_x = MS_NINT(marker_width/2.0);
                marker_offset_y = MS_NINT(marker_height/2.0);

                marker_rect.minx = MS_NINT(cachePtr->point.x - .5 * marker_width);
                marker_rect.miny = MS_NINT(cachePtr->point.y - .5 * marker_height);
                marker_rect.maxx = marker_rect.minx + (marker_width-1);
                marker_rect.maxy = marker_rect.miny + (marker_height-1);

                /* actually draw the marker */
                if (layerPtr->type == MS_LAYER_ANNOTATION) draw_marker = 1;
            }

            if (labelPtr->position == MS_AUTO)
            {
                if (layerPtr->type == MS_LAYER_LINE)
                {
                    int position = MS_UC;
            
                    /* Two angles or two positions, depending on angle. Steep angles */
                    /* will use the angle approach, otherwise we'll rotate between*/
                    /* UC and LC. */
                    for (j=0; j<2; j++)
                    {
                        msFreeShape(cachePtr->poly);
                        /* assume label *can* be drawn */
                        cachePtr->status = MS_TRUE;

                        p = get_metrics(&(cachePtr->point), position, r,
                                        (marker_offset_x + labelPtr->offsetx),
                                        (marker_offset_y + labelPtr->offsety),
                                        labelPtr->angle, labelPtr->buffer, cachePtr->poly);

                        /*save marker bounding polygon*/
                        if (draw_marker)
                            msRectToPolygon(marker_rect, cachePtr->poly);

                        /* Compare against image bounds, rendered labels and markers
                           (sets cachePtr->status) */
                        msTestLabelCacheCollisions(&(map->labelcache), labelPtr, 
                                                   map->width, map->height, 
                                                   labelPtr->buffer, cachePtr, priority, l);

                        /*found a suitable place for this label*/
                        if (cachePtr->status)
                            break;

                    }  /*next angle*/

                }
                else
                {
                    /* loop through the outer label positions */
                    for (j=0; j<=7; j++)
                    {
                        msFreeShape(cachePtr->poly);
                        /* assume label *can* be drawn */
                        cachePtr->status = MS_TRUE;

                        p = get_metrics(&(cachePtr->point), j, r,
                                        (marker_offset_x + labelPtr->offsetx),
                                        (marker_offset_y + labelPtr->offsety),
                                        labelPtr->angle, labelPtr->buffer, cachePtr->poly);

                        /* save marker bounding polygon*/
                        if (draw_marker)
                            msRectToPolygon(marker_rect, cachePtr->poly);

                        /* Compare against image bounds, rendered labels and markers
                           (sets cachePtr->status) */
                        msTestLabelCacheCollisions(&(map->labelcache), labelPtr, 
                                                   map->width, map->height, 
                                                   labelPtr->buffer, cachePtr, priority, l);

                        /* found a suitable place for this label*/
                        if (cachePtr->status)
                            break;
                    } /*next position*/
                }

                /* draw in spite of collisions based on last position,
                   need a *best* position */
                if (labelPtr->force) cachePtr->status = MS_TRUE;
            }
            else
            {
                /* assume label *can* be drawn */
                cachePtr->status = MS_TRUE;

                /*don't need the marker_offset*/
                if (labelPtr->position == MS_CC)
                {
                    p = get_metrics(&(cachePtr->point), labelPtr->position, r,
                                    labelPtr->offsetx, labelPtr->offsety,
                                    labelPtr->angle,  labelPtr->buffer, cachePtr->poly);
                }
                else
                {
                    p = get_metrics(&(cachePtr->point), labelPtr->position, r,
                                    (marker_offset_x + labelPtr->offsetx),
                                    (marker_offset_y + labelPtr->offsety),
                                    labelPtr->angle, labelPtr->buffer, cachePtr->poly);
                }

                /* save marker bounding polygon, part of overlap tests */
                if (draw_marker)
                    msRectToPolygon(marker_rect, cachePtr->poly);

                if (!labelPtr->force)
                { /* no need to check anything else*/
                    /* Compare against image bounds, rendered labels and markers
                       (sets cachePtr->status) */
                    msTestLabelCacheCollisions(&(map->labelcache), labelPtr, 
                                               map->width, map->height, 
                                               labelPtr->buffer, cachePtr, priority, l);
                }
            } /* end position if-then-else */


            /* next label */
            if (!cachePtr->status)
                continue;

            /* need to draw a marker */
            if (layerPtr->type == MS_LAYER_ANNOTATION && cachePtr->numstyles > 0)
            {
                for (i=0; i<cachePtr->numstyles; i++)
                {
                    msDrawMarkerSymbolPDF(&map->symbolset, image, &(cachePtr->point),
                                          &(cachePtr->styles[i]), layerPtr->scalefactor);
                }
            }

            /*handle a filled label background*/
            /* TODO */
            /* if(labelPtr->backgroundcolor >= 0) */
            /* billboardPDF(img, cachePtr->poly, labelPtr); */

            msDrawTextPDF(image, p, cachePtr->text,
                          labelPtr, &(map->fontset), layerPtr->scalefactor);

        } /* next in cache */
    } /* next priority */

    return(0);
}

/************************************************************************/
/*  void msDrawShadeSymbolPDF                                           */
/*                                                                      */
/*  Draw a shade symbol of the specified size and color                 */
/************************************************************************/
void msDrawShadeSymbolPDF(symbolSetObj *symbolset, imageObj *image,
                          shapeObj *p, styleObj *style, double scalefactor)
{
    int              size;
    MAPPDF_DOC_TYPE  pdf;

    if (style->size == -1)
    {
        size = msSymbolGetDefaultSize(symbolset->symbol[style->symbol]);
        size = MS_NINT(size*scalefactor);
    }
    else
        size = MS_NINT(style->size*scalefactor);
    pdf = image->img.pdf->pdf;

    if (p->numlines <= 0)
        return;
    /* no such symbol, 0 is OK */
    if (style->symbol > symbolset->numsymbols || style->symbol < 0)
        return;

    if (size < 1) /* size too small */
        return;

    if (MS_VALID_COLOR(style->color))
        msFilledPolygonPDF(pdf, p, &(style->color));

    if (MS_VALID_COLOR(style->outlinecolor))
        drawPolylinePDF(pdf, p, &(style->outlinecolor), size);

    return;
}


/************************************************************************/
/*  void msDrawMarkerSymbolPDF                                          */
/*                                                                      */
/*  Draw a single marker symbol of the specified size and color         */
/************************************************************************/
void msDrawMarkerSymbolPDF(symbolSetObj *symbolset, imageObj *image,
                           pointObj *p, styleObj *style, double scalefactor)
{
    symbolObj        *symbol;
    MAPPDF_DOC_TYPE  pdf;
#if defined USE_HARUPDF
    HPDF_Page        page = NULL;
#endif
    int              offset_x, offset_y, x, y;
    int              j;
    char             symbolBuffer[2];
    int              size;
    double           scale = 1.0;

    pdf = image->img.pdf->pdf;
#if defined USE_HARUPDF
    page = HPDF_GetCurrentPage(pdf);
    if (!page)
    {
        msSetError(MS_GDERR, "Unable to get current pdf page.", "msDrawMap()");
        return;
    }
#endif

    /* set the colors */
    /* if no outline color, make the stroke and fill the same */
    if (!(MS_VALID_COLOR(style->outlinecolor))) style->outlinecolor=style->color;

    MAPPDF_FN_SETRGBSTROKECOLOR(MAPPDF_DRAWOBJ,
                                (float)(style->outlinecolor.red/255),
                                (float)(style->outlinecolor.green/255),
                                (float)(style->outlinecolor.blue/255));
    MAPPDF_FN_SETRGBFILLCOLOR(MAPPDF_DRAWOBJ, (float)(style->color.red/255),
                                              (float)(style->color.green/255),
                                              (float)(style->color.blue/255));

    /*  set up the symbol scale and type */
    symbol = symbolset->symbol[style->symbol];
    if (style->size == -1)
    {
        size = msSymbolGetDefaultSize( symbol );
        size = MS_NINT(size*scalefactor);
    }
    else
        size = MS_NINT(style->size * scalefactor);

    size = MS_MAX(size, style->minsize);
    size = MS_MIN(size, style->maxsize);

    /* no such symbol, 0 is OK */
    if (style->symbol > symbolset->numsymbols || style->symbol < 0)
        return;
    if (size < 1) /* size too small */
        return;

/* -------------------------------------------------------------------- */
/*      Render the diffrent type of symbols.                            */
/* -------------------------------------------------------------------- */
    switch (symbol->type)
    {
/* -------------------------------------------------------------------- */
/*      Symbol : true type.                                             */
/* -------------------------------------------------------------------- */
        case (MS_SYMBOL_TRUETYPE):
        {
#if defined USE_HARUPDF
            char       *font = NULL;
            HPDF_Font  font_id;
#elif defined USE_PDFLIB
            int        font_id;
#endif

            sprintf(symbolBuffer, "%c", (char)*symbol->character);

#if defined USE_HARUPDF
                font = msLookupHashTable(&(image->img.pdf->hfonts), symbol->font);
                font_id = HPDF_GetFont(pdf, font, NULL);
                free(font);
#elif defined USE_PDFLIB
                font_id = PDF_findfont(pdf, symbol->font ,"winansi", 1);
#endif

            MAPPDF_FN_SETFONT(MAPPDF_DRAWOBJ, font_id, size+2);
#if defined USE_HARUPDF
            x = p->x - (int)(.5 * HPDF_Page_TextWidth(page, symbolBuffer));
#elif defined USE_PDFLIB
            /**** FIXME  was this a mistake? font size set +2, but stringwidth
                         was requested for +0 (changed to +2 now) ****/
            x = p->x - (int)(.5 * PDF_stringwidth(pdf, symbolBuffer, font_id, size+2));
#endif
            y = p->y;
            MAPPDF_FN_SETLINEWIDTH(MAPPDF_DRAWOBJ, 0.15);
            MAPPDF_Scale(MAPPDF_DRAWOBJ, 1, -1);
#if defined USE_HARUPDF
            HPDF_Page_BeginText(page);
#endif
            MAPPDF_FN_TEXTXY(MAPPDF_DRAWOBJ, symbolBuffer, x, -y);
#if defined USE_HARUPDF
            HPDF_Page_EndText(page);
#endif
            MAPPDF_Scale(MAPPDF_DRAWOBJ, 1, -1);

            break;
        }
/* -------------------------------------------------------------------- */
/*      Symbol : pixmap.                                                */
/* -------------------------------------------------------------------- */
        case (MS_SYMBOL_PIXMAP):
        {
            int         length;
#if defined USE_HARUPDF
            HPDF_BYTE   *data;
            HPDF_Image  result;
#elif defined USE_PDFLIB
            char        *data;
            int         result;
#endif
            float       imageScale = 1.0;

/*                FILE *pngOut;
            char tempFile[32];
            tempFile[0]=0;

            sprintf(tempFile,"/tmp/%d.png",getpid());
            msSetError(MS_MISCERR, tempFile, "drawPng()");
            pngOut = fopen(tempFile,"wb");
            gdImagePng(symbol->img,pngOut);
            style->colorlose(pngOut);
            result = PDF_open_image_file(pdf,"png",tempFile,"",0);
*/
            if (size>10 && symbol->img->sx > size)
            {
                imageScale = (float)((float)size / (float)symbol->img->sx);
            }

            length = 0;
#if defined USE_HARUPDF
            /* requires at least haru 2.2.0 */
            data = gdImagePngPtr(symbol->img, &length);
            result = HPDF_LoadPngImageFromMem(pdf, data, (HPDF_UINT32)length);
#elif defined USE_PDFLIB
            /**** FIXME  use PDFlib virtual files to load PNG from mem
                         avail v5, check for PDF_load_image() in configure ****/
            data = gdImageJpegPtr(symbol->img, &length, 95);
            result = PDF_open_image(pdf, "jpeg", "memory",
                                    data, (long)length,
                                    symbol->img->sx, symbol->img->sy,
                                    3, 8, NULL);
#endif
            gdFree(data);

            MAPPDF_Scale(MAPPDF_DRAWOBJ, 1, -1);
#if defined USE_HARUPDF
            HPDF_Page_DrawImage(page, result, p->x-symbol->img->sx / 2 * imageScale,
                                -p->y-symbol->img->sy / 2 * imageScale,
                                symbol->img->sx * imageScale, symbol->img->sy * imageScale);
#elif defined USE_PDFLIB
            PDF_place_image(pdf, result, p->x-symbol->img->sx / 2 * imageScale,
                            -p->y-symbol->img->sy / 2 * imageScale, imageScale);
#endif
            MAPPDF_Scale(MAPPDF_DRAWOBJ, 1, -1);

#if defined USE_PDFLIB
            PDF_close_image(pdf, result);
#endif
        }
        break;
/* -------------------------------------------------------------------- */
/*      symbol : Ellipse                                                */
/* -------------------------------------------------------------------- */
        case (MS_SYMBOL_ELLIPSE):
                /* ok, this is going to be just a circle */

            scale = size / symbol->sizey;
            x = MS_NINT(symbol->sizex * scale);
            y = MS_NINT(symbol->sizey * scale);

            MAPPDF_FN_CIRCLE(MAPPDF_DRAWOBJ, p->x, p->y, (x)/2);

            if (symbol->filled)
                MAPPDF_FN_FILLSTROKE(MAPPDF_DRAWOBJ);
            else
                MAPPDF_FN_STROKE(MAPPDF_DRAWOBJ);

            break;
/* -------------------------------------------------------------------- */
/*      symbol : Vector.                                                */
/* -------------------------------------------------------------------- */
        case (MS_SYMBOL_VECTOR):

            scale = size / symbol->sizey;
            /* set the joins to be mitered - better for small shapes */
            MAPPDF_FN_SETLINEJOIN(MAPPDF_DRAWOBJ, MAPPDF_LINEJOIN_MITER);

            offset_x = MS_NINT(p->x - scale*.5*symbol->sizex);
            offset_y = MS_NINT(p->y - scale*.5*symbol->sizey);

            MAPPDF_FN_MOVETO(MAPPDF_DRAWOBJ,
                       MS_NINT(scale*symbol->points[0].x + offset_x),
                       MS_NINT(scale*symbol->points[0].y + offset_y));
            for (j = 0; j < symbol->numpoints; j++)
            {
                if (MS_NINT(symbol->points[j].x >= 0) &&
                    MS_NINT(symbol->points[j].y) >= 0)
                {
                    if (MS_NINT(symbol->points[j-1].x < 0) &&
                        MS_NINT(symbol->points[j-1].y < 0))
                    {
                        MAPPDF_FN_MOVETO(MAPPDF_DRAWOBJ,
                                   MS_NINT(scale*symbol->points[j].x + offset_x),
                                   MS_NINT(scale*symbol->points[j].y + offset_y));
                    }
                    else
                    {
                        MAPPDF_FN_LINETO(MAPPDF_DRAWOBJ,
                                   MS_NINT(scale*symbol->points[j].x + offset_x),
                                   MS_NINT(scale*symbol->points[j].y + offset_y));
                    }
                }
            }

            if(symbol->filled)
            { /* if filled */
                MAPPDF_FN_CLOSEFILLSTROKE(MAPPDF_DRAWOBJ);
            }
            else
            { /* NOT filled */
                MAPPDF_FN_STROKE(MAPPDF_DRAWOBJ);
            }
            /* set the joins back to rounded */
            MAPPDF_FN_SETLINEJOIN(MAPPDF_DRAWOBJ, MAPPDF_LINEJOIN_ROUND);
        
            break;
        default:
            break;
    } /* end switch statement */

    return;
}

/************************************************************************/
/*  int msDrawRasterLayerPDF                                            */
/*                                                                      */
/*  adds a raster image to the pdf.                                     */
/************************************************************************/
int msDrawRasterLayerPDF(mapObj *map, layerObj *layer, imageObj *image)
{
    outputFormatObj  *format = NULL;
    imageObj         *image_tmp = NULL;
    MAPPDF_DOC_TYPE  pdf = NULL;
#if defined USE_HARUPDF
    HPDF_Page        page = NULL;
#endif
    int              nLength = 0;
#if defined USE_HARUPDF
    HPDF_BYTE        *data = NULL;
    HPDF_Image       nResult;
#elif defined USE_PDFLIB
    char             *data = NULL;
    int              nResult = 0;
#endif
    int              bRasterOutput = 0;

    assert(strcasecmp(map->outputformat->driver, "PDF") == 0);
    pdf = image->img.pdf->pdf;

#if defined USE_HARUPDF
    page = HPDF_GetCurrentPage(pdf);
    if (!page)
    {
        msSetError(MS_GDERR, "Unable to get current pdf page.", "msDrawMap()");
        return(-1);
    }
#endif

/* -------------------------------------------------------------------- */
/*      create a temporary GD image and render in it.                   */
/* -------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/*  Idea is to render the image into image_tmp using the ms GD api,      */
/*  use gd to convert it to a png. PDFlib can then place the png into the*/
/*  output pdf. Done as jpeg for now - change.                            */
/* --------------------------------------------------------------------- */
    format = msCreateDefaultOutputFormat( NULL, "GD/PC256" );
    if (format == NULL)
        return -1;

    bRasterOutput = 0;
    if (strcasecmp(msGetOutputFormatOption(image->format, "OUTPUT_TYPE", ""), 
                   "RASTER") == 0)
    {
        image_tmp = (imageObj *)image->img.pdf->imagetmp;
        bRasterOutput = 1;
    }
    else
        image_tmp = msImageCreate(image->width, image->height, format,
                                  NULL, NULL, map);

    if (image_tmp == NULL)
        return -1;


    if (msDrawRasterLayerLow(map, layer, image_tmp) != -1)
    {
/* -------------------------------------------------------------------- */
/*      if it is a RASTER output, just return. At save time the         */
/*      temporary image will be output to the pdf object.               */
/* -------------------------------------------------------------------- */
        if (bRasterOutput)
            return 0;

#if defined USE_HARUPDF
        /* requires at least haru 2.2.0 */
        data = gdImagePngPtr(image_tmp->img.gd, &nLength);
        nResult = HPDF_LoadPngImageFromMem(pdf, data, (HPDF_UINT32)nLength);
#elif defined USE_PDFLIB
        /*kludge: this should really be a raw image or png. jpeg is not good*/
        /*but pdflib doesn't support in mem opening of png.*/
        /**** FIXME  use PDFlib virtual files to load PNG from mem
                     avail v5, check for PDF_load_image() in configure ****/
        data = gdImageJpegPtr(image_tmp->img.gd, &nLength, 95);
        nResult = PDF_open_image(pdf, "jpeg", "memory", data, (long)nLength,
                                map->width, map->height, 3, 8, NULL);
#endif
        gdFree(data);

        MAPPDF_FN_GSAVE(MAPPDF_DRAWOBJ); /* save the original coordinate system */
        MAPPDF_Scale(MAPPDF_DRAWOBJ, 1, -1); /* flip the coordinates, and therefore the image */
#if defined USE_HARUPDF
        HPDF_Page_DrawImage(page, nResult, 0, -(map->height),
                            map->width, map->height);
#elif defined USE_PDFLIB
        PDF_place_image(pdf, nResult, 0, -(map->height), 1.0);
#endif
        MAPPDF_FN_GRESTORE(MAPPDF_DRAWOBJ); /* restore the original coordinate system */

#if defined USE_PDFLIB
        PDF_close_image(pdf,nResult);
#endif
        msFreeImage(image_tmp);

    }
    else
    {
        msSetError(MS_GDERR, "couldn't covert a raster layer to GD image",
                   "msDrawRasterLayerPDF()");
        msFreeImage(image_tmp);
        return -1;
    }
    
    return 0;
}


/************************************************************************/
/*  int msDrawVectorLayerAsRasterPDF                                    */
/*                                                                      */
/*  draw vectors to image and add it to the PDF at save time.           */
/************************************************************************/
int msDrawVectorLayerAsRasterPDF(mapObj *map, layerObj *layer, imageObj*image)
{
    imageObj *imagetmp;

    if (!image || !MS_DRIVER_PDF(image->format))
      return MS_FAILURE;


    if (strcasecmp(msGetOutputFormatOption(image->format, "OUTPUT_TYPE",""), 
                   "RASTER") != 0)
    {
        return MS_FAILURE;
    }

    imagetmp = (imageObj *)image->img.pdf->imagetmp;

    if (imagetmp)
    {
        msImageInitGD(imagetmp, &map->imagecolor);
        /* msLoadPalette(imagetmp->img.gd, &(map->palette), map->imagecolor); */
        msDrawVectorLayer(map, layer, imagetmp);
        
        return MS_SUCCESS;
    }

    return MS_FAILURE;
}


/************************************************************************/
/*  void msTransformShapePDF                                            */
/*                                                                      */
/*  Transform geographic coordinates to output coordinates.             */
/*                                                                      */
/************************************************************************/
void msTransformShapePDF(shapeObj *shape, rectObj extent, double cellsize)
{
    int i,j;

    if (shape->numlines == 0) return;

    if (shape->type == MS_SHAPE_LINE || shape->type == MS_SHAPE_POLYGON)
    {
        for (i = 0; i < shape->numlines; i++)
        {
            for (j = 0; j < shape->line[i].numpoints; j++ )
            {
                shape->line[i].point[j].x =
                    (shape->line[i].point[j].x - extent.minx)/cellsize;
                shape->line[i].point[j].y =
                    (extent.maxy - shape->line[i].point[j].y)/cellsize;
            }
        }
        return;
    }
}

/************************************************************************/
/*  void msSaveImagePDF                                                 */
/*                                                                      */
/*  writes the image's pdf out to disk or sends it to stdout             */
/************************************************************************/
int msSaveImagePDF(imageObj *image, char *filename)
{

    if (image && MS_DRIVER_PDF(image->format))
    {
        FILE             *stream;
#if defined USE_HARUPDF
        HPDF_UINT32      size;
        HPDF_BYTE        *pdfBuffer;
#elif defined USE_PDFLIB
        long             size;
        const char       *pdfBuffer;
#endif
        MAPPDF_DOC_TYPE  pdf = NULL;
#if defined USE_HARUPDF
        HPDF_Page        page = NULL;
#endif
        int              nLength = 0;
#if defined USE_HARUPDF
        HPDF_BYTE        *data = NULL;
        HPDF_Image       nResult;
#elif defined USE_PDFLIB
        char             *data = NULL;
        int              nResult = 0;
#endif
        imageObj         *imagetmp;

        mapObj *map = image->img.pdf->map;

        if (strcasecmp(msGetOutputFormatOption(image->format,"OUTPUT_TYPE",""), 
                       "RASTER") == 0)
        {
            /* FILE *out; */

            pdf = image->img.pdf->pdf;

#if defined USE_HARUPDF
            page = HPDF_GetCurrentPage(pdf);
            if (!page)
            {
                msSetError(MS_GDERR, "Unable to get current pdf page.", "msDrawMap()");
                return(-1);
            }
#endif

            imagetmp = (imageObj *)image->img.pdf->imagetmp;
#if defined USE_HARUPDF
            /* requires at least haru 2.2.0 */
            data = gdImagePngPtr(imagetmp->img.gd, &nLength);
            nResult = HPDF_LoadPngImageFromMem(pdf, data, (HPDF_UINT32)nLength);
#elif defined USE_PDFLIB
            /**** FIXME  use PDFlib virtual files to load PNG from mem
                         avail v5, check for PDF_load_image() in configure ****/
            data = gdImageJpegPtr(imagetmp->img.gd, &nLength, 95);
            nResult = PDF_open_image(pdf, "jpeg", "memory", data, (long)nLength,
                                     map->width, map->height, 3, 8, NULL);
#endif
            gdFree(data);

            MAPPDF_FN_GSAVE(MAPPDF_DRAWOBJ); /* save the original coordinate system */
            MAPPDF_Scale(MAPPDF_DRAWOBJ, 1, -1); /* flip the coordinates, and therefore the image */
#if defined USE_HARUPDF
            HPDF_Page_DrawImage(page, nResult, 0, -(map->height),
                                map->width, map->height);
#elif defined USE_PDFLIB
            PDF_place_image(pdf,nResult, 0, -(map->height), 1.0);
#endif
            MAPPDF_FN_GRESTORE(MAPPDF_DRAWOBJ); /* restore the original coordinate system */

#if defined USE_PDFLIB
            PDF_close_image(pdf,nResult);
#endif
            msFreeImage(image->img.pdf->imagetmp);
        }
        /*finish the page and put the entire pdf into a buffer*/
        /* haru saves files and streams, takes more work to get buffers */
        /* can easily save direct to file:
            HPDF_SaveToFile (image->img.pdf->pdf, filename);
        */
#if defined USE_HARUPDF
        HPDF_STATUS status = HPDF_SaveToStream(image->img.pdf->pdf);
        size = HPDF_GetStreamSize(image->img.pdf->pdf);
        HPDF_ResetStream(pdf);
        pdfBuffer = (HPDF_BYTE *)malloc(sizeof(HPDF_BYTE) * size);
        status = HPDF_ReadFromStream(image->img.pdf->pdf, &pdfBuffer[0], &size);
#elif defined USE_PDFLIB
        PDF_end_page(image->img.pdf->pdf);
        PDF_close(image->img.pdf->pdf);
        pdfBuffer = PDF_get_buffer(image->img.pdf->pdf, &size);
#endif
        if (filename != NULL && strlen(filename) > 0)
        {
            stream = fopen(filename, "wb");
            if (!stream)
            {
                msSetError(MS_IOERR, "(%s)", "msSaveImagePDF()", filename);
                return(MS_FAILURE);
            }
        }
        else
        { /* use stdout */
            /* Change stdout mode to binary on win32 platforms*/
            if (msIO_needBinaryStdout() == MS_FAILURE)
                return MS_FAILURE;
            stream = stdout;
        }

      /*-----------------------------------------------
      /send the active buffer to disk
      /-----------------------------------------------*/

      msIO_fwrite(pdfBuffer, sizeof(char), size, stream);
      if (filename != NULL && strlen(filename) > 0)
         fclose(stream);
      /* free(pdfBuffer); */

      return(MS_SUCCESS);
    }
    msSetError(MS_MISCERR, "Incorrect driver passed as pdf: %s.",
                     "msSaveImagePDF()", image->format );
    return MS_FAILURE;
}

/************************************************************************/
/*  void msFreeImagePDF                                                 */
/*                                                                      */
/*  Free pdf object structure                                           */
/************************************************************************/
void msFreeImagePDF(imageObj *image)
{
    if (image && MS_DRIVER_PDF(image->format) &&
        image->img.pdf && image->img.pdf->pdf)
    {
        MAPPDF_FREE(image->img.pdf->pdf);
        image->img.pdf->pdf = NULL;
    }
}

/************************************************************************/
/*  int msDrawTextPDF                                                   */
/*                                                                      */
/*  creates a text element in the pdf                                   */
/************************************************************************/
int msDrawTextPDF(imageObj *image, pointObj labelPnt, char *string,
                 labelObj *label, fontSetObj *fontset, double scalefactor)
{
    int              x, y, x1, y1;
#if defined USE_HARUPDF
    HPDF_Font        font = NULL;
#elif defined USE_PDFLIB
    int              font = 0;
#endif
    int              size;
    float            phi = label->angle;
    colorObj         sColor;
    char             *wrappedString;
    char             *fontPath = NULL;
    char             *fontOutlineAlias = NULL;
#if defined USE_PDFLIB
    int              nFontOutlineLen = 0;
#endif
    MAPPDF_DOC_TYPE  pdf;
#if defined USE_HARUPDF
    HPDF_Page        page = NULL;
#endif

/* -------------------------------------------------------------------- */
/*      if not PDF, return.                                             */
/* -------------------------------------------------------------------- */
    if (image == NULL || !MS_DRIVER_PDF(image->format))
        return 0;

/* -------------------------------------------------------------------- */
/*      validate arguments.                                             */
/* -------------------------------------------------------------------- */
    if (!string || strlen(string) == 0 || !label || !fontset)
        return(0); /* not an error, just don't want to do anything */

    if (strlen(string) == 0)
        return(0); /* not an error, just don't want to do anything */

    x = MS_NINT(labelPnt.x);
    y = MS_NINT(labelPnt.y);

    size = label->size * scalefactor;

    if (!fontset)
    {
        msSetError(MS_TTFERR, "No fontset defined.", "msDrawTextPDF()");
        return(-1);
    }

    if (!label->font)
    {
        msSetError(MS_TTFERR, "No font defined.", "msDrawTextPDF()");
        return(-1);
    }

    fontPath = msLookupHashTable(&(fontset->fonts), label->font);

    if (!fontPath)
    {
        msSetError(MS_TTFERR, "Requested font (%s) not found.", "msDrawTextPDF()",
                   label->font);
        return(-1);
    }

    pdf = image->img.pdf->pdf;

#if defined USE_HARUPDF
    page = HPDF_GetCurrentPage(pdf);
    if (!page)
    {
        msSetError(MS_GDERR, "Unable to get current pdf page.", "msDrawMap()");
        return(-1);
    }
#endif

/* -------------------------------------------------------------------- */
/* Do color initialization stuff                                        */
/* -------------------------------------------------------------------- */
    sColor.red = 0;
    sColor.green = 0;
    sColor.blue = 0;

    if (MS_VALID_COLOR(label->color))
    {
        sColor.red = label->color.red;
        sColor.green = label->color.green;
        sColor.blue = label->color.blue;
    }
    else if (MS_VALID_COLOR(label->outlinecolor))
    {
        sColor.red = label->outlinecolor.red;
        sColor.green = label->outlinecolor.green;
        sColor.blue = label->outlinecolor.blue;
    }
    else if (MS_VALID_COLOR(label->shadowcolor))
    {
        sColor.red = label->shadowcolor.red;
        sColor.green = label->shadowcolor.green;
        sColor.blue = label->shadowcolor.blue;
    }
    else
    {
        msSetError(MS_TTFERR, "Invalid color", "draw_textPDF()");
        return(-1);
    }

    if (!string)
        return(0); /* not an error, just don't want to do anything */

    if (strlen(string) == 0)
        return(0); /* not an error, just don't want to do anything */

    x = MS_NINT(labelPnt.x);
    y = MS_NINT(labelPnt.y);
    x1 = x; y1 = y;

    MAPPDF_FN_SETRGBSTROKECOLOR(MAPPDF_DRAWOBJ, (float)sColor.red/255,
                                                (float)sColor.green/255,
                                                (float)sColor.blue/255);
    MAPPDF_FN_SETRGBFILLCOLOR(MAPPDF_DRAWOBJ, (float)sColor.red/255,
                                              (float)sColor.green/255,
                                              (float)sColor.blue/255);
    MAPPDF_FN_SETLINEWIDTH(MAPPDF_DRAWOBJ, .3);

    /*set up font handling*/
#if defined USE_HARUPDF
    fontOutlineAlias = msLookupHashTable(&(image->img.pdf->hfonts), label->font);
    font = HPDF_GetFont(pdf, fontOutlineAlias, NULL);
    free(fontOutlineAlias);
#elif defined USE_PDFLIB
    /*set an alias for the font outline file*/
    /**** FIXME  why use fontpaths here when fonts were loaded into PDF in 
                 msLoadFontSetPDF() ? ****/
    nFontOutlineLen = strlen(fontPath);
    fontOutlineAlias = (char *)malloc(sizeof(char)*(nFontOutlineLen+4));
    sprintf(fontOutlineAlias, "f1=%s", fontPath);
    PDF_set_parameter(pdf, "FontOutline", fontOutlineAlias);
    free(fontOutlineAlias);

    /*load font using alias*/
    font = PDF_load_font(pdf, "f1", 0, "winansi", NULL);
#endif
    MAPPDF_FN_SETFONT(MAPPDF_DRAWOBJ, font, size+2);

#if defined USE_HARUPDF
    HPDF_Page_BeginText(page);
#endif

    if (phi != 0){
        /*PDF_save(pdf);*/
        MAPPDF_Translate(MAPPDF_DRAWOBJ, x, y);
        MAPPDF_Rotate(MAPPDF_DRAWOBJ, -phi);
        x = y = 0;
    }

    MAPPDF_Scale(MAPPDF_DRAWOBJ, 1, -1);
    if ((wrappedString = strchr(string, '\r')) == NULL)
    {
        MAPPDF_FN_TEXTXY(MAPPDF_DRAWOBJ, string, x, -y);
    }
    else
    {
        char *headPtr;
        headPtr = string;
            /* break the string into pieces separated by \r\n */
        while (wrappedString){
            char *piece;
            int length = wrappedString - headPtr;
            piece = malloc(length+1);
            memset(piece, 0, length+1);
            strncpy(piece, headPtr, length);

            if (headPtr == string){
                MAPPDF_FN_TEXTXY(MAPPDF_DRAWOBJ, piece, x, -y);
            }
            else {
                MAPPDF_FN_TEXTNEXTLINE(MAPPDF_DRAWOBJ, piece);
            }
            free(piece);
            headPtr = wrappedString + 2;
            wrappedString = strchr(headPtr, '\r');
        }
        MAPPDF_FN_TEXTNEXTLINE(MAPPDF_DRAWOBJ, headPtr);
    }
    MAPPDF_Scale(MAPPDF_DRAWOBJ, 1, -1);
    if (phi != 0){
        MAPPDF_Rotate(MAPPDF_DRAWOBJ, phi);
        MAPPDF_Translate(MAPPDF_DRAWOBJ, -x1, -y1);
/* PDF_restore(pdf); */
    }
#if defined USE_HARUPDF
    HPDF_Page_EndText(page);
#endif
    MAPPDF_FN_SETLINEWIDTH(MAPPDF_DRAWOBJ, 1);
    return 0;
}

/************************************************************************/
/*  void msDrawStartShapePDF                                            */
/*                                                                      */
/*  setup to start placing shapes in the pdf                            */
/************************************************************************/
void msDrawStartShapePDF(mapObj *map, layerObj *layer, imageObj *image,
                         shapeObj *shape)
{
    return;
}

/************************************************************************/
/*  void msImageStartLayerPDF                                           */
/*                                                                      */
/*  set up to start drawing a layer                                     */
/************************************************************************/
void msImageStartLayerPDF(mapObj *map, layerObj *layer, imageObj *image)
{
    return;
}

/************************************************************************/
/*  int msEmbedScalebarPDF                                              */
/*                                                                      */
/*  draw a scalebar into the pdf.                                       */
/************************************************************************/
int msEmbedScalebarPDF(mapObj *map, imageObj *image)
{
    /*TODO*/
    return(0);
}

/************************************************************************/
/*                          int msDrawWMSLayerPDF                       */
/*                                                                      */
/*      Use low level gd functions to draw a wms layer in a gd          */
/*      images                                                          */
/************************************************************************/
int msDrawWMSLayerPDF(int nLayerId, httpRequestObj *pasReqInfo, 
                      int numRequests, mapObj *map, layerObj *layer, imageObj *image)
{
    MAPPDF_DOC_TYPE     pdf = NULL;
#if defined USE_HARUPDF
    HPDF_Page           page = NULL;
#endif
    imageObj            *image_tmp = NULL;
    int                 iReq = -1;
    char                *driver = strdup("GD/GIF");
    int                 nLength = 0;
#if defined USE_HARUPDF
    HPDF_BYTE           *data = NULL;
    HPDF_Image          nResult;
#elif defined USE_PDFLIB
    char                *data = NULL;
    int                 nResult = 0;
#endif
    /* char                ttt[200]; */

    int                 bRasterOutput = 0;

#ifdef USE_GD_GIF
    driver = strdup("GD/GIF");
#else

#ifdef USE_GD_PNG
     driver = strdup("GD/PNG");
#else

#ifdef USE_GD_JPEG
     driver = strdup("GD/JPEG");
#else

#ifdef USE_GD_WBMP
     driver = strdup("GD/WBMP");
#endif

#endif
#endif
#endif


    if (!image || !MS_DRIVER_PDF(image->format) || image->width <= 0 ||
        image->height <= 0)
        return -1;

    pdf = image->img.pdf->pdf;

#if defined USE_HARUPDF
    page = HPDF_GetCurrentPage(pdf);
    if (!page)
    {
        msSetError(MS_GDERR, "Unable to get current pdf page.", "msDrawMap()");
        return(-1);
    }
#endif


/* ==================================================================== */
/*      WMS requests are done simultaniously for all WMS layer.         */
/* ==================================================================== */
    for(iReq=0; iReq<numRequests; iReq++)
    {
        if (pasReqInfo[iReq].nLayerId == nLayerId)
          break;
    }
    if (iReq == numRequests)
      return MS_SUCCESS;
/* -------------------------------------------------------------------- */
/*      create a temprary GD image and render in it.                    */
/* -------------------------------------------------------------------- */
    if (strcasecmp(msGetOutputFormatOption(image->format,
                                           "OUTPUT_TYPE",""), 
                      "RASTER") != 0)
    {
        image_tmp = msImageCreateGD(map->width, map->height,
                                    msCreateDefaultOutputFormat(map, driver),
                                    map->web.imagepath, map->web.imageurl);
    }
    else
    {
        image_tmp = (imageObj *)image->img.pdf->imagetmp;
        bRasterOutput = 1;
    }


    msImageInitGD(image_tmp, &map->imagecolor);

    if (msDrawWMSLayerLow(nLayerId, pasReqInfo, numRequests, map, layer, 
                          image_tmp) != -1)
    {
        if (bRasterOutput)
            return MS_SUCCESS;

#if defined USE_HARUPDF
        /* requires at least haru 2.2.0 */
        data = gdImagePngPtr(image_tmp->img.gd, &nLength);
        nResult = HPDF_LoadPngImageFromMem(pdf, data, (HPDF_UINT32)nLength);
#elif defined USE_PDFLIB
        /*kludge: this should really be a raw image or png. jpeg is not good*/
        /*but pdflib doesn't support in mem opening of png.*/
        /**** FIXME  use PDFlib virtual files to load PNG from mem
                     avail v5, check for PDF_load_image() in configure ****/
        data = gdImageJpegPtr(image_tmp->img.gd, &nLength, 95);
        nResult = PDF_open_image(pdf, "jpeg", "memory", data, (long)nLength,
                                map->width, map->height, 3, 8, NULL);
#endif

        gdFree(data);

        MAPPDF_FN_GSAVE(MAPPDF_DRAWOBJ); /* save the original coordinate system */
        MAPPDF_Scale(MAPPDF_DRAWOBJ, 1, -1); /* flip the coordinates, and therefore the image */
#if defined USE_HARUPDF
        HPDF_Page_DrawImage(page, nResult, 0, -(map->height),
                            map->width, map->height);
#elif defined USE_PDFLIB
        PDF_place_image(pdf, nResult, 0, -(map->height), 1.0);
#endif
        MAPPDF_FN_GRESTORE(MAPPDF_DRAWOBJ); /* restore the original coordinate system */

#if defined USE_PDFLIB
        PDF_close_image(pdf,nResult);
#endif
        msFreeImage(image_tmp);
    }

    return MS_SUCCESS;
}

#endif
